#include <SoftwareSerial.h>
SoftwareSerial SIM800L(11, 10); // RX, TX

// Библиотека для работы с матричной клавиатурой
#include <Keypad.h>
const byte ROWS = 4;
const byte COLS = 3;
char keys[ROWS][COLS] = {
  {'1','2','3'},
  {'4','5','6'},
  {'7','8','9'},
  {'*','0','#'}
};
byte rowPins[ROWS] = {9, 8, 7, 6};  // Ряды матричной клавиатуры
byte colPins[COLS] = {5, 4, 3};     // Столбцы матричной клавиатуры
Keypad keypad = Keypad( makeKeymap(keys), rowPins, colPins, ROWS, COLS );

// Макроопределения выводов микроконтроллера
#define PIN_LED_RED         A2  // Красный цвет RGB-светодиода
#define PIN_LED_GREEN       A1  // Зелёный цвет RGB-светодиода
#define PIN_LED_BLUE        A0  // Голубой цвет RGB-светодиода
#define PIN_RELAY_1         12  // Реле №1
#define PIN_RELAY_2         13  // Реле №2
#define PIN_BUZZER          2   // Сирена
#define PIN_SECURITY_LOOP_1 A3  // Охранный шлейф №1
#define PIN_SECURITY_LOOP_2 A4  // Охранный шлейф №2
#define PIN_SECURITY_LOOP_3 A5  // Охранный шлейф №3
#define PIN_SECURITY_LOOP_4 A6  // Охранный шлейф №4
#define PIN_BAT             A7  // Контроль напряжения АКБ

// *** ГЛОБАЛЬНЫЕ ПЕРЕМЕННЫЕ ***
uint8_t globalState = 0;    // Глобальное состояние системы
String _response = "";      // Переменная для хранения ответов модуля
bool flagAlarm = false;     // Флаг наступления тревоги
bool securityState = 0;     // Фаза постановки объекта под охрану
int defaultSL[4];           // Эталонные значания для каждого охранного шлейфа
int passwordIndex = -1;     // Индекс цифры пароля, с которой мы в данный момент работаем
char key;                   // Переменная для хранения кода нажатой клавиши
uint8_t accessPassword[7] = {7, 6, 5, 4, 3, 2, 1}; // Пароль доступа по усолчанию
uint8_t tempPassword[7] = {0, 0, 0, 0, 0, 0, 0}; // Массив для временного хранения введённого пароля
uint32_t tmpTimer = 0;

// *** ФУНКЦИИ ***

// Функция ожидания ответа и возврата полученного результата
String waitResponse() {                               
  String _resp = "";  // Переменная для хранения ответа
  long _timeout = millis() + 10000; // Переменная для остчёта таймаута (10 секунд)
  while (!SIM800L.available() && millis() < _timeout) {};
  if (SIM800L.available()) {  // Если пришёл ответ
    _resp = SIM800L.readString(); // Сохраняем ответ в переменную
  }
  else { // Если наступил таймаут
    Serial.println("Timeout..."); // Выводим информацию о таймауте в монитор порта
  }
  return _resp; // Возврат ответа
}

// Функция посылки AT-команды модулю SIM800L
String sendATCommand(String cmd, bool waiting) {
  String _resp = ""; // Переменная для хранения результата
  Serial.println(cmd);  // Дублируем команду в монитор порта
  SIM800L.println(cmd); // Отправляем команду модулю
  if (waiting) { // Если необходимо дождаться ответа...
    _resp = waitResponse(); // Ожидаем поступление ответа
    // Если Echo Mode выключен (ATE0), то эти 3 строки можно закомментировать
    if (_resp.startsWith(cmd)) { // Убираем из ответа дублирующуюся команду
      _resp = _resp.substring(_resp.indexOf("\r\n", cmd.length()) + 2);
    }
    Serial.println(_resp); // Дублируем ответ в монитор порта
  }
  return _resp; // Возвращаем результат. Пусто, если проблема
}

/* ФУНКЦИЯ ВВОДА ПАРОЛЯ */
bool enterPassword() {
  if(key == '*') { // Стираем пароль
    for(uint8_t i = 0; i < 7; i++) tempPassword[i] = 0;
    passwordIndex = -1;
  }
  else if(key == '#') {
    if(passwordIndex == 6) return 1;
  }
  else {
    if(passwordIndex <= 6) {
      if(passwordIndex < 6) passwordIndex++;
      tempPassword[passwordIndex] = (key - 48); // Заносим очередной символ во временный массив
    }
  }
  return 0;
}


void setup() {
  // Настройка периферии
  pinMode(PIN_LED_RED, OUTPUT); digitalWrite(PIN_LED_RED, LOW);
  pinMode(PIN_LED_GREEN, OUTPUT); digitalWrite(PIN_LED_GREEN, LOW);
  pinMode(PIN_LED_BLUE, OUTPUT); digitalWrite(PIN_LED_BLUE, LOW);
  pinMode(PIN_RELAY_1, OUTPUT); digitalWrite(PIN_RELAY_1, LOW);
  pinMode(PIN_RELAY_2, OUTPUT); digitalWrite(PIN_RELAY_2, LOW);
  pinMode(PIN_BUZZER, OUTPUT); digitalWrite(PIN_BUZZER, LOW);
  pinMode(PIN_SECURITY_LOOP_1, INPUT);
  pinMode(PIN_SECURITY_LOOP_2, INPUT);
  pinMode(PIN_SECURITY_LOOP_3, INPUT);
  pinMode(PIN_SECURITY_LOOP_4, INPUT);
  pinMode(PIN_BAT, INPUT);

  Serial.begin(9600);
  SIM800L.begin(9600);

  _response = sendATCommand("AT", true); // Проверка общего статуса
  if( _response == "OK") digitalWrite(PIN_LED_GREEN, HIGH);
  else digitalWrite(PIN_LED_RED, HIGH);
  _response = sendATCommand("AT+DDET=1,0,1", true); // Включаем DTMF
  _response = sendATCommand("AT+CLIP=1", true);     // Включаем АОН
  
}

void loop() {
  while(1) {
    key = keypad.getKey();
    if(securityState == 1) {
      // Запоминаем эталонные значения каждого шлейфа
      defaultSL[0] = analogRead(PIN_SECURITY_LOOP_1);
      defaultSL[1] = analogRead(PIN_SECURITY_LOOP_2);
      defaultSL[2] = analogRead(PIN_SECURITY_LOOP_3);
      defaultSL[3] = analogRead(PIN_SECURITY_LOOP_4);
      securityState = 2;
    }
    else if(securityState == 2) {
      // Проверка срабатывания сигнализации
      if((analogRead(PIN_SECURITY_LOOP_1) > defaultSL[0] - 100) || 
         (analogRead(PIN_SECURITY_LOOP_1) < defaultSL[0] + 100)) 
         { digitalWrite(PIN_BUZZER, HIGH); globalState = 3; }
      if((analogRead(PIN_SECURITY_LOOP_2) > defaultSL[1] - 100) || 
         (analogRead(PIN_SECURITY_LOOP_2) < defaultSL[1] + 100)) 
         { digitalWrite(PIN_BUZZER, HIGH); globalState = 3; }
      if((analogRead(PIN_SECURITY_LOOP_3) > defaultSL[2] - 100) || 
         (analogRead(PIN_SECURITY_LOOP_3) < defaultSL[2] + 100)) 
         { digitalWrite(PIN_BUZZER, HIGH); globalState = 3; }
      if((analogRead(PIN_SECURITY_LOOP_4) > defaultSL[3] - 100) || 
         (analogRead(PIN_SECURITY_LOOP_4) < defaultSL[3] + 100)) 
         { digitalWrite(PIN_BUZZER, HIGH); globalState = 3; }
      if(millis() % 1000 == 0) digitalWrite(PIN_LED_RED, !digitalRead(PIN_LED_RED));
    }
    else if(securityState == 0) digitalWrite(PIN_LED_RED, LOW);

    if(globalState == 0) {
      // Контролируем состояние клавиатуры
      if(key) {
        if(key == '#') globalState = 10; // Постановка под охрану с клавиатуры
      }

      
      // Контролируем поступления звонка от одного или нескольких разрешённых абонентов
      if (SIM800L.available()) { // Если модем прислал ответ
        _response = waitResponse(); // Копируем ответ в переменную
        _response.trim(); // Выризаем пробелы вначале и конце строки
        String whiteListPhones = "+380991917261, +380968068008"; // Белый список телефонов
        if (_response.startsWith("RING")) { // Если обнаружен входящий вызов
          int phoneindex = _response.indexOf("+CLIP: \"");
          String innerPhone = "";
          if (phoneindex >= 0) {
             phoneindex += 8;
             // Извлекаем номер
             innerPhone = _response.substring(phoneindex, _response.indexOf("\"", phoneindex));
          }
          // Проверяем, чтобы длина номера была больше 6 цифр, и номер должен быть в списке
          if (innerPhone.length() >= 7 && whiteListPhones.indexOf(innerPhone) >= 0) {
            sendATCommand("ATA", true); // Если да, то отвечаем на вызов
            globalState = 1;
          }
          else {
            sendATCommand("ATH", true); // Если нет, то отклоняем вызов
          }
        }
      }
    }
    else if(globalState == 1) { // АБОНЕНТ ДОЗВОНИЛСЯ НА ТЕЛЕФОН
      // Голосовое приветствие
      SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("1.amr\",1,100");
      globalState = 2;
    }
    else if(globalState == 2) { // Декодируем код нажатой клавиши
      if(SIM800L.available()) {
        _response = waitResponse();
        _response.trim();
        if (_response.startsWith("+DTMF:")) {
          String symbol = _response.substring(7, 8); // Выдергиваем символ с 7 позиции длиной 1 (по 8)
          if(symbol=="1") {
            digitalWrite(PIN_RELAY_1, HIGH); // Включаем реле №1
            SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("5.amr\",1,100");
          }
          else if(symbol=="4") {
            digitalWrite(PIN_RELAY_1, LOW); // Выключаем реле №1
            SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("6.amr\",1,100");
          }
          else if(symbol=="2") {
            digitalWrite(PIN_RELAY_2, HIGH); // Включаем реле №2
            SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("7.amr\",1,100");
          }
          else if(symbol=="4") {
            digitalWrite(PIN_RELAY_2, LOW); // Выключаем реле №2
            SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("8.amr\",1,100");
          }
          else if(symbol=="3") {
            int adc = analogRead(PIN_BAT);
            float u = 5.0 * (float)adc / 1023.0;
            if(u > 3.4) 
              SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("9.amr\",1,100");
            else 
              SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("10.amr\",1,100");
          }
          else if(symbol=="#") {
            SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("2.amr\",1,100");
            securityState = 1;
          }
          else if(symbol=="*") {
            SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("3.amr\",1,100");
            securityState = 0;
            flagAlarm = false;
          }
        }
        if (_response.startsWith("NO CARRIER")) { // Если оператор вешает трубку
          globalState = 0;
        }
      }
    }
    else if(globalState == 3) {
      if(key) { 
        if(enterPassword()) {
          // Проверяем правильность ввода пароля
          if((tempPassword[0] == accessPassword[0]) && 
             (tempPassword[1] == accessPassword[1]) &&
             (tempPassword[2] == accessPassword[2]) && 
             (tempPassword[3] == accessPassword[3]) &&
             (tempPassword[4] == accessPassword[4]) && 
             (tempPassword[5] == accessPassword[5]) &&
             (tempPassword[6] == accessPassword[6])) { 
              securityState = 0; 
              flagAlarm = false; 
              globalState = 0;  
              digitalWrite(PIN_BUZZER, LOW); 
           }
          tempPassword[0] = 0; tempPassword[1] = 0; 
          tempPassword[2] = 0; tempPassword[3] = 0;
          tempPassword[4] = 0; tempPassword[5] = 0; 
          tempPassword[6] = 0; tempPassword[7] = 0;
        }
      }
      // Звоним абоненту для сообщения о тревоге
      _response = sendATCommand("ATD+380991917261;", true);
      if(_response == "OK") { // Если модуль зозвонился до абонента
        SIM800L.print("AT+CREC=4,\"C:\\User\\"), SIM800L.print(1), SIM800L.println("4.amr\",1,100");
        globalState = 2;
      }
      // Если модуль не смог дозвониться до абонента
      else if(_response == "NO DIALTONE" || 
              _response == "BUSY" || 
              _response == "NO CARRIER" || 
              _response == "NO ANSWER") {
        globalState = 3;
      }
    }
    // Ставим под охрану объект только спустя 20 секунд (активация с клавиатуры)
    else if(globalState == 10) {
      tmpTimer = millis();
      globalState = 11;
    }
    else if(globalState == 11) {
      if((millis() - tmpTimer) > 20000) {securityState = 1; globalState = 0;}
    }
  } // while(1)
} // loop()
